--[[=============================================================================
    Get, Handle and Dispatch message functions

    Copyright 2015 Control4 Corporation. All Rights Reserved.
===============================================================================]]

-- This macro is utilized to identify the version string of the driver template version used.
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.device_messages = "2015.03.31"
end

--[[=============================================================================
    GetMessage()
  
    Description:
    Used to retrieve a message from the communication buffer. Each driver is
    responsible for parsing that communication from the buffer.
  
    Parameters:
    None
  
    Returns:
    A single message from the communication buffer
===============================================================================]]
function GetMessage()

    if (gReceiveBuffer:len() == 0) then return "" end

    local message, pos, ack, nack
    local pattern = ""
    
	--pattern = "^(" .. gACKHeader .. "%z%z%z." .. ")()"
    pattern = "(" .. gACKHeader .. "%z%z%z." .. ")()"
    ack, pos = gReceiveBuffer:match(pattern)
    if (ack) then
	   message = ack
	   gReceiveBuffer = gReceiveBuffer:sub(pos)
	   return message
    end
    
	--pattern = "^(%z" .. tohex('14') .. "%z%z%z." .. ")()
    pattern = "(%z" .. tohex('14') .. "%z%z%z." .. ")()"
    nack, pos = gReceiveBuffer:match(pattern)
    if (nack) then
	   message = nack
	   gReceiveBuffer = gReceiveBuffer:sub(pos)
	   return message
    end
    
    --pattern = "^" .. gReadResponseHeader .. ".(.)"
    pattern = gReadResponseHeader .. ".(.)"
    local len = gReceiveBuffer:match(pattern)
    if (len) then
	   _, len = string.unpack(len, "b")
	   if (len == 3) then
		  if (gReceiveBuffer:len() < 9) then return "" end
		  pattern = gReadResponseHeader .. ".....(.).()"
	   elseif (len == 4) then
		  if (gReceiveBuffer:len() < 10) then return "" end
		  pattern = gReadResponseHeader .. ".....(..).()"
	   elseif (len == 6) then	--lamp usage
			 if (gReceiveBuffer:len() < 12) then return "" end
			 pattern = gReadResponseHeader .. ".....(....).()"
	   elseif (len == 10) then	--operating temperature
		  if (gReceiveBuffer:len() < 16) then return "" end
		  pattern = gReadResponseHeader .. ".....(....).....()"	
	   elseif (len == 22) then	--error packet...
		  if (gReceiveBuffer:len() < 28) then return "" end
		  pattern = gReadResponseHeader .. ".....(....................).()"	
	   else
		  LogWarn("GetMessage(), Invalid message length: " .. len)
		  dbgHexdump(gReceiveBuffer)
		  return ""
	   end
	   message, pos = gReceiveBuffer:match(pattern)
    else
	   LogWarn("GetMessage(), Null message length: ")
	   dbgHexdump(gReceiveBuffer)
	   return ""
    end

    gReceiveBuffer = gReceiveBuffer:sub(pos)
    
    if (gReceiveBuffer:len() > 0) then
	   LogWarn("GetMessage(), Discarding remnant from buffer: ")
	   dbgHexdump(gReceiveBuffer)
    end
    gReceiveBuffer = ""
    
    return message

end

--[[=============================================================================
    HandleMessage(message)]

    Description
    This is where we parse the messages returned from the GetMessage()
    function into a command and data. The call to 'DispatchMessage' will use the
    'name' variable as a key to determine which handler routine, function, should
    be called in the DEV_MSG table. The 'value' variable will then be passed as
    a string parameter to that routine.

    Parameters
    message(string) - Message string containing the function and value to be sent to
                      DispatchMessage

    Returns
    Nothing
===============================================================================]]
function HandleMessage(message)
    if (type(message) == "table") then
	   LogTrace("HandleMessage. Message is ==>")
	   LogTrace(message)
	   LogTrace("<==")
    else
	   LogTrace("HandleMessage. Message is:")
	   dbgHexdump(message)
    end

    local last_command = gCon._LastCommand.command_name
    LogInfo('last_command: ' .. tostring(last_command))
    
    if (message == gNACK_Packet) then
	   LogTrace("HandleMessage." .. last_command .. " command is currently disabled!")
	   
	   if (last_command == "PULSE_VOL_UP" or last_command == "PULSE_VOL_DOWN") then
		  --do nothing, fail gracefully...volume may already be at min or max 
	   elseif (last_command == "MUTE_ON" or last_command == "MUTE_OFF" or last_command == "MUTE_TOGGLE") then
		  --do nothing, fail gracefully
		  if (last_command == "MUTE_TOGGLE") then
			 --commands may fail if mute state is the same as the mute command that was just sent
			 if (gMuteState == "False") then
				gMuteState = "True"
			 else
				gMuteState = "False"
			 end
		  end
	   else
		  if ((last_command == "ON") and (gCon._CommandRetryCount == 0)) then -- unit amy already be ON...
			 gQueryPowerDelayTimer:StartTimer(3, "SECONDS")
		  end
			  
		  if ((last_command == "CURRENT_INPUT") and (gCon._CommandRetryCount == 1)) then
			 C4:FireEvent("Input Selection Fail")
		  end
			 
		  return
	   end
    end    
    
    gCon._CommandRetryCount = 0
    gCon._WaitResponseTimer:KillTimer()
    local command_delay
    if (gProjectorProxy._VolumeIsRamping == true) then
	   command_delay = tonumber(Properties["Volume Ramp Delay Milliseconds"])
    else
	   command_delay = tonumber(Properties["Command Delay Milliseconds"])
    end
    gHandleACKDelayTimer:StartTimer(command_delay, "MILLISECONDS")
    
    if (message == gACK_Packet) then
	   if (last_command == "ON") then
		  gQueryPowerDelayTimer:StartTimer(tonumber(Properties["Power Status Query Delay Seconds"]), "SECONDS")
		  
		  gQueryLampHoursDelayTimer:StartTimer(5, "MINUTES")
		  
	   elseif (last_command == "OFF") then
		  gQueryPowerDelayTimer:StartTimer(tonumber(Properties["Power Status Query Delay Seconds"]), "SECONDS")
	   elseif (last_command == "SET_VOLUME_LEVEL" or last_command == "PULSE_VOL_DOWN" or last_command == "PULSE_VOL_UP") then
		  gQueryVolumeDelayTimer:StartTimer(tonumber(Properties["Volume Ramp Delay Milliseconds"]), "MILLISECONDS")
		  gQueryVolumeDelayTimer._CallbackParam = "Volume" 
	   elseif (last_command == "MUTE_OFF" or last_command == "MUTE_ON" or last_command == "MUTE_TOGGLE") then
		  gQueryVolumeDelayTimer:StartTimer(500, "MILLISECONDS")
		  gQueryVolumeDelayTimer._CallbackParam = "Mute" 	   
	   elseif (last_command == "SET_INPUT") then
		  gQueryInputDelayTimer:StartTimer(6, "SECONDS")
	   end
	   
	   return
	   
    end
    
    local name = last_command or ""
    local value = message or ""

    DispatchMessage(name, value)
	
end

function OnHandleACKDelayTimerExpired()
    LogTrace("OnHandleACKDelayTimerExpired")
    gHandleACKDelayTimer:KillTimer()
    
    gCon:HandleACK()
end

function OnQueryVolumeDelayTimerExpired(command)
    LogTrace("OnQueryVolumeDelayTimerExpired")
    gQueryVolumeDelayTimer:KillTimer()
    
    if (command == "Volume") then
	   PackAndQueueReadCommand("VOLUME_STATUS", CMDS.GET_VOLUME_STATUS)
    elseif (command == "Mute") then
	   PackAndQueueReadCommand("MUTE_STATUS", CMDS.GET_MUTE_STATUS)
    else
	   PackAndQueueReadCommand("VOLUME_STATUS", CMDS.GET_VOLUME_STATUS)
	   PackAndQueueReadCommand("MUTE_STATUS", CMDS.GET_MUTE_STATUS)
    end
end
    
function OnQueryPowerDelayTimerExpired()
    LogTrace("OnQueryPowerDelayTimerExpired")
    gQueryPowerDelayTimer:KillTimer()
       
    PackAndQueueReadCommand("POWER_STATUS", CMDS.GET_POWER_STATUS)
    
    --if gPowerOnTimer:TimerStarted() then
	   gQueryPowerDelayTimer:StartTimer(tonumber(Properties["Power Status Query Delay Seconds"]), "SECONDS")
    --end
    
end

function OnPowerOnTimerExpired()
    LogTrace("OnPowerOnTimerExpired")
    gPowerOnTimer:KillTimer()
    
    gQueryPowerDelayTimer:StartTimer(tonumber(2), "SECONDS")
    
    if gSetInputDelayTimer:TimerStarted() then
		  gSetInputDelayTimer:StartTimer(4, "SECONDS")
    end
		  
    gQueryVolumeDelayTimer:StartTimer(8, "SECONDS")
    gQueryVolumeDelayTimer._CallbackParam = "Volume_Mute" 
end

function OnPowerOffTimerExpired()
    LogTrace("OnPowerOffTimerExpired")
    gPowerOffTimer:KillTimer()
    
    --for quick power on after power off...
    if gPowerOnTimer:TimerStarted() then
	   gSendPowerOnAfterPowerOffDelayTimer:StartTimer(2, "SECONDS")
    end
end

function OnSendPowerOnAfterPowerOffDelayTimerExpired()
    LogTrace("OnSendPowerOnAfterPowerOffDelayTimerExpired")
    gSendPowerOnAfterPowerOffDelayTimer:KillTimer()
    gCheckPowerStateFromOnCommand = false
    gPowerOnTimer:StartTimer(tonumber(Properties["Power On Delay Seconds"]), "SECONDS")    
    ON()
end

function OnQueryLampHoursDelayTimerExpired()
    LogTrace("OnQueryLampHoursDelayTimerExpired")
    gQueryLampHoursDelayTimer:KillTimer()
    
    PackAndQueueReadCommand("LAMP_HOURS", CMDS.GET_LAMP_HOURS)
    
    gQueryLampHoursDelayTimer:StartTimer(60, "MINUTES")
end


--[[=============================================================================
    DispatchMessage(MsgKey, MsgData)

    Description
    Parse routine that will call the routines to handle the information returned
    by the connected system.

    Parameters
    MsgKey(string)  - The function to be called from within DispatchMessage
    MsgData(string) - The parameters to be passed to the function found in MsgKey

    Returns
    Nothing
===============================================================================]]
function DispatchMessage(MsgKey, MsgData)
	if (DEV_MSG[MsgKey] ~= nil and (type(DEV_MSG[MsgKey]) == "function")) then
		LogInfo("DEV_MSG." .. tostring(MsgKey))
		--dbgHexdump(MsgData)
		local status, err = pcall(DEV_MSG[MsgKey], MsgData)
		if (not status) then
			LogError("LUA_ERROR: " .. err)
		end
	else
		LogTrace("HandleMessage: Unhandled command = " .. MsgKey)
	end
end

--[[
TODO: Create DEV_MSG functions for all messages to call Notifies.
	  Sample functions are included below for all applicable notifications.
--]]

function DEV_MSG.CURRENT_INPUT(value)
     _, value = string.unpack(value, "b")
	LogTrace("DEV_MSG.CURRENT_INPUT(), value = " .. value)
	local input = tInputConnMapByName[tInputResponseMap[value]].ID 	--mod 1000 value of Input Connection ID
     local bindingID = tInputConnMapByID[input].BindingID 		
	LogInfo("INPUT_CHANGED, input = " .. tInputResponseMap[value])
	
    --[[
    NOTE: Updated v2.10 (bug has existed for a long time but this is my current understanding...)
    There is TV proxy bug that affects Current Input in Composer Programming. 
    If connection has audio then it is considered audio, ONLY send audio connection notification.
    For video only connections (i.e. HDMI), BOTH audio and video connection notifications must be sent.
    --]]
    
    --all connections
    gProjectorProxy:dev_InputChanged(bindingID, input + 3000)		--audio
    
    --video only connections
    if (input > 1) then
	   gProjectorProxy:dev_InputChanged(bindingID, input + 1000)		--video
    end
end

function DEV_MSG.POWER_STATUS(value)
     _, value = string.unpack(value, "b")
	LogTrace("DEV_MSG.POWER_STATUS(), value = " .. value)

	--[[
	TODO: "00" & "01" & "02" & "03" & "04" values will need to be edited based upon  
		the device protocol values, indicating if the device is "off", "on", 
		in a transitional state (i.e. Warming up, Cooling down) or in an error State.
		Then modify the if/then code block below to match the desired states.
	--]]
	if (value == 0) then
	     if (gCheckPowerStateFromOnCommand) then
		  LogInfo("Checking Power State From ON Command: Projector is OFF, sending ON command.")
		  gCheckPowerStateFromOnCommand = false
		  --gPowerOnTimer:StartTimer(tonumber(Properties["Power On Delay Seconds"]), "SECONDS")
		  --PackAndQueueCommand("ON", CMDS["ON"])
	   
		  ----------
		  local command_delay = tonumber(Properties["Power On Delay Seconds"])
		  local delay_units = "SECONDS"
		  
		  if gPowerOffTimer:TimerStarted() then
			 --quick power on AFTER power off, need TV to turn off before sending on. This occurs in OnPowerOffTimer...
			 local off_time_remaining = tonumber(Properties["Power Off Delay Seconds"]) - (os.time() - gPowerOffStartTime)
			 LogInfo("Delaying Power ON until Power OFF sequence has finished! off_time_remaining: " .. off_time_remaining)
			 gPowerOnTimer:StartTimer(command_delay + off_time_remaining + 4, delay_units)
			 return
		  else
			 gPowerOnTimer:StartTimer(command_delay, delay_units)
		  end
		  
		  --sometimes needs two to wake up...
		  --PackAndQueueCommand("ON", CMDS["ON"])
		  PackAndQueueCommand("ON", CMDS["ON"])
		  ----------
		  return
		end
	   
		if (gProjector_Operating_Status ~= "OFF") then
			gProjector_Operating_Status = "OFF"
			C4:SetVariable('Projector_Operating_Status',"OFF")
			gProjectorProxy:dev_PowerOff()
			gQueryPowerDelayTimer:KillTimer()
		end
				
	elseif (value== 1) then
	     if (gPowerOffTimer:TimerStarted()) then return end
		  
	     if (gCheckPowerStateFromOnCommand) then
		  LogInfo("Checking Power State From ON Command: projector is already ON.")
		  gCheckPowerStateFromOnCommand = false
		  gPowerOnTimer:KillTimer()
		  
		  ----------
		  if gSetInputDelayTimer:TimerStarted() then
			 gSetInputDelayTimer:StartTimer(1, "SECONDS")
		  end
		  
		  if (gProjector_Operating_Status ~= "ON") then
			gProjector_Operating_Status = "ON"
			C4:SetVariable('Projector_Operating_Status',"ON")
			gProjectorProxy:dev_PowerOn()
		  end	
		  
		  --gQueryVolumeDelayTimer:StartTimer(4, "SECONDS")
		  --gQueryVolumeDelayTimer._CallbackParam = "Volume_Mute" 
		  ----------
		  
		  return
		end
		
		if (gProjector_Operating_Status ~= "ON") then
			gProjector_Operating_Status = "ON"
			C4:SetVariable('Projector_Operating_Status',"ON")
			gProjectorProxy:dev_PowerOn()
		end	
		gQueryPowerDelayTimer:KillTimer()
		
		
    --[[
	elseif (value== "02") then	
		if (gProjector_Operating_Status ~= "COOLING_DOWN") then
			gProjector_Operating_Status = "COOLING_DOWN"
			C4:SetVariable('Projector_Operating_Status',"COOLING_DOWN")
		end
		local command_delay = tonumber(Properties["Power Status Query Delay Seconds"])
		local delay_units = "SECONDS"
		gProjectorProxy._PowerStatusQueryDelayTimer:StartTimer(command_delay, delay_units)	
	elseif (value== "03") then	
		if (gProjector_Operating_Status ~= "WARMING_UP") then
			gProjector_Operating_Status = "WARMING_UP"
			C4:SetVariable('Projector_Operating_Status',"WARMING_UP")
		end
		local command_delay = tonumber(Properties["Power Status Query Delay Seconds"])
		local delay_units = "SECONDS"
		gProjectorProxy._PowerStatusQueryDelayTimer:StartTimer(command_delay, delay_units)	
	elseif (value== "04") then	
		if (gProjector_Operating_Status ~= "ERROR_CONDITION") then
			gProjector_Operating_Status = "ERROR_CONDITION"
			C4:SetVariable('Projector_Operating_Status',"ERROR_CONDITION")
		end
		--TODO: Handle the error...
		-- most likely want to send an Off Notification to the proxy
		-- so the power on will work after the error condition is satisfied
		gProjectorProxy:dev_PowerOff()
    --]]
	else
		LogWarn("DEV_MSG.POWER(): value not valid - " .. value)
	end		
end

function DEV_MSG.LAMP_HOURS(value)
     _, value = string.unpack(value, "h")
	LogTrace("Projector reports " .. value .. " Hours on the bulb")	
	C4:UpdateProperty('Lamp Hours', tostring(value))
	if (gLamp_Hours ~= value) then
		gLamp_Hours = value
		C4:SetVariable('Lamp_Hours',tostring(value))
	end	
end
	
function DEV_MSG.VOLUME_STATUS(value)
     _, value = string.unpack(value, "b")
	LogTrace("DEV_MSG.VOLUME_STATUS(), value = " .. value)
	
	-- TODO: derive and set  "output" from value or create separate DEV_MSG functions for each Output Connection
	local output = 0 --mod 1000 value of Output Connection ID
	
	local deviceLevel = value
	local minDeviceLevel = MIN_DEVICE_LEVEL
	local maxDeviceLevel = MAX_DEVICE_LEVEL
	--C4 volume level uses a percentage scale: 0 - 100
	local c4Level = ConvertVolumeToC4(deviceLevel, minDeviceLevel, maxDeviceLevel)		
	
	gProjectorProxy:dev_VolumeLevelChanged(output, c4Level, deviceLevel)		
end

function DEV_MSG.MUTE_STATUS(value)
     _, value = string.unpack(value, "b")
	LogTrace("DEV_MSG.MUTE_STATUS(), value = " .. value)
	
	local output = 0    --mod 1000 value of Output Connection ID
	
	local state   	--Mute state represented as "True" or "False"
	-- TODO: values "01" & "00" will need to be modified based upon the device protocol specification
	if (value == 1) then 
		state = "True"
	elseif (value == 0) then
		state = "False"
	else
		LogWarn("DEV_MSG.MUTE(), value not valid, exiting...")
		return
	end
	
	if (gMuteToggleInProgress) then
	     gMuteToggleInProgress = false
		if (state == "True") then
		  LogTrace("MUTE_TOGGLE: MUTE_OFF command")
		  PackAndQueueCommand("MUTE_TOGGLE", CMDS["MUTE_OFF"])
	     else
		  LogTrace("MUTE_TOGGLE: MUTE_ON command")
		  PackAndQueueCommand("MUTE_TOGGLE", CMDS["MUTE_ON"])
		end
	
	else
	    gProjectorProxy:dev_MuteChanged(output, state)
	    gMuteState = state
	end
end		

function DEV_MSG.BASS(value)
	LogTrace("DEV_MSG.BASS(), value = " .. value)
	
	-- TODO: derive and set  "output" from value or create separate DEV_MSG functions for each Output Connection
	local output = 0  --mod 1000 value of Output Connection ID
	
	-- TODO: set "level", Bass level is represented as a percentage value
	local level  
	
	gProjectorProxy:dev_BassLevelChanged(output, level)
end	

function DEV_MSG.TREBLE(value)
	LogTrace("DEV_MSG.TREBLE(), value = " .. value)
	
	-- TODO: derive and set  "output" from value or create separate DEV_MSG functions for each Output Connection
	local output = 0 --mod 1000 value of Output Connection ID
	
	-- TODO: set "level", Treble level is represented as a percentage value
	local level   
	
	gProjectorProxy:dev_TrebleLevelChanged(output, level)
end	

function DEV_MSG.BALANCE(value)
	LogTrace("DEV_MSG.BALANCE(), value = " .. value)
	
	-- TODO: derive and set  "output" from value or create separate DEV_MSG functions for each Output Connection
	local output = 0  --mod 1000 value of Output Connection ID
	
	-- TODO: set "level", Bass level is represented as a percentage value
	local level  
	
	gProjectorProxy:dev_BalanceLevelChanged(output, level)
end	

function DEV_MSG.LOUDNESS(value)
	LogTrace("DEV_MSG.LOUDNESS(), value = " .. value)
	
	-- TODO: derive and set  "output" from value or create separate DEV_MSG functions for each Output Connection
	local output = 0  --mod 1000 value of Output Connection ID
	
	-- TODO: set "state", Loudness state is represented as "True" or "False" (literal string, not boolean)
	local state  
	gProjectorProxy:dev_LoudnessChanged(output, state)
end
	
