--[[=============================================================================
    Basic Template for PROJECTOR Driver

    Copyright 2015 Control4 Corporation. All Rights Reserved.
===============================================================================]]------------
require "common.c4_driver_declarations"
require "common.c4_common"
require "common.c4_init"
require "common.c4_property"
require "common.c4_command"
require "common.c4_notify"
require "common.c4_network_connection"
require "common.c4_serial_connection"
require "common.c4_ir_connection"
require "common.c4_utils"
require "lib.c4_timer"
require "actions"
require "device_specific_commands"
require "device_messages"
require "projector_init"
require "properties"
require "proxy_commands"
require "connections"
require "projector.projector_proxy_class"
require "projector.projector_proxy_commands"
require "projector.projector_proxy_notifies"


-- This macro is utilized to identify the version string of the driver template version used.
if (TEMPLATE_VERSION ~= nil) then
	TEMPLATE_VERSION.driver = "2015.03.31"
end

--[[=============================================================================
    Constants
===============================================================================]]
PROJECTOR_PROXY_BINDINGID = 5001

--[[
    Device Volume Range
    TODO: edit "MIN_DEVICE_LEVEL" & "MAX_DEVICE_LEVEL" values based upon the protocol specification for volume range.
    If zones have volume ranges that vary, convert these constants into a table indexed by the mod value of the output connection 
    and then update all references to handle the table structure.
--]]
MIN_DEVICE_LEVEL = 0
MAX_DEVICE_LEVEL = 20	


--[[=============================================================================
    Initialization Code
===============================================================================]]
function ON_DRIVER_EARLY_INIT.main()
	
end

function ON_DRIVER_INIT.main()

	-- TODO: If cloud based driver then uncomment the following line
	--ConnectURL()
end

function ON_DRIVER_LATEINIT.main()
	C4:urlSetTimeout(20)
	DRIVER_NAME = C4:GetDriverConfigInfo("name")
	
	SetLogName(DRIVER_NAME)
end

function ON_DRIVER_EARLY_INIT.projector_driver()

end

function ON_DRIVER_INIT.projector_driver()
    -- TODO: Modify tVolumeRamping to have on entry per Output connection
    --index is mod 1000 value of output connection
    local tVolumeRamping = {["state"] = false, ["mode"] = ""}
	
    -- Create an instance of the PROJECTOR Proxy class
    -- TODO: Change bProcessesDeviceMessages to false if Device Messages will not be processes
    local  bProcessesDeviceMessages = true
    local bUsePulseCommandsForVolumeRamping = false
    if (Properties["Ramping Method"] == "Pulse Commands") then bUsePulseCommandsForVolumeRamping = true end
    gProjectorProxy = ProjectorProxy:new(PROJECTOR_PROXY_BINDINGID,  bProcessesDeviceMessages, tVolumeRamping, bUsePulseCommandsForVolumeRamping)
    gProjectorProxy:Initialize()
    
    --timers
    gQueryLampHoursDelayTimer = c4_timer:new("QueryLampHoursDelayTimer", 60, "MINUTES", OnQueryLampHoursDelayTimerExpired)
    gQueryPowerDelayTimer = c4_timer:new("QueryPowerDelayTimer", 1, "SECONDS", OnQueryPowerDelayTimerExpired)
    gPowerOnTimer = c4_timer:new("PowerOnTimer", tonumber(Properties["Power On Delay Seconds"]), "SECONDS", OnPowerOnTimerExpired)
    gPowerOffTimer = c4_timer:new("PowerOffTimer", tonumber(Properties["Power Off Delay Seconds"]), "SECONDS", OnPowerOffTimerExpired)   
    gSendPowerOnAfterPowerOffDelayTimer = c4_timer:new("SendPowerOnAfterPowerOffDelayTimer", tonumber(Properties["Power Off Delay Seconds"]), "SECONDS", OnSendPowerOnAfterPowerOffDelayTimerExpired)
    gSetInputDelayTimer = c4_timer:new("SetInputDelayTimer", 2, "SECONDS", OnSetInputDelayTimerExpired, command)
    gQueryInputDelayTimer = c4_timer:new("QueryInputDelayTimer", 2, "SECONDS", OnQueryInputDelayTimerExpired)
    gQueryVolumeDelayTimer = c4_timer:new("QueryVolumeDelayTimer", 1, "SECONDS", OnQueryVolumeDelayTimerExpired, command)
    gVolumeRampDelayTimer = c4_timer:new("VolumeRampDelayTimer", tonumber(Properties["Power On Delay Seconds"]), "SECONDS", OnVolumeRampDelayTimerExpired)    
    gHandleACKDelayTimer = c4_timer:new("HandleACKDelayTimer", 500, "MILLISECONDS", OnHandleACKDelayTimerExpired)
    
    gWriteFunctionCommandHeader = tohex('06 14 00 04 00 34')
    gWriteFunctionCommandHeader_RemoteKey = tohex('02 14 00 04 00 34')
    gReadFunctionCommandHeader = tohex('07 14 00 05 00 34 00 00')
    gACK_Packet = tohex('03 14 00 00 00 14')
    gNACK_Packet = tohex('00 14 00 00 00 14')
    gReadResponseHeader = tohex('05 14')
    gACKHeader = tohex('03 14')
    gNACKHeader = tohex('00 14')
    
    gMuteState = ""
    
    local minDeviceLevel = MIN_DEVICE_LEVEL
    local maxDeviceLevel = MAX_DEVICE_LEVEL
    tVolumeCurve = getVolumeCurve(minDeviceLevel, maxDeviceLevel)
	
	--[[
	For the "Volume Curve" method, tVolumeCurve is used to store volume level values that will be used to build volume commands during volume ramping. Specifically, they are used in GetNextVolumeCurveValue() which is called from the ContinueVolumeRamping() function.  Property values for "Volume Ramping Steps" and "Volume Ramping Slope" can be adjusted to get a smooth volume ramping from low to high volume.
	--]]
end

function getVolumeCurve(minDeviceLevel, maxDeviceLevel)
    local steps = tonumber(Properties["Volume Ramp Steps"])
    local slope = tonumber(Properties["Volume Ramp Slope"])
    local tCurve = gProjectorProxy:CreateVolumeCurve(steps, slope, minDeviceLevel, maxDeviceLevel)
    
    return tCurve 
end

function ON_DRIVER_LATEINIT.projector_driver()
	--C4:SetPropertyAttribs ("Command Delay Milliseconds", 1)
	C4:SetPropertyAttribs ("== Volume Ramping Settings ==", 1)
	C4:SetPropertyAttribs ("Ramping Method", 1)
	C4:SetPropertyAttribs ("Volume Ramp Steps", 1)
	C4:SetPropertyAttribs ("Volume Ramp Slope", 1)
end

--[[=============================================================================
    Driver Code
===============================================================================]]
function PackAndQueueCommand(...)
    local command_name = select(1, ...) or ""
    local command = select(2, ...) or ""
    local command_delay = select(3, ...) or tonumber(Properties["Command Delay Milliseconds"])
    local delay_units = select(4, ...) or "MILLISECONDS"
    LogTrace("PackAndQueueCommand(), command_name = " .. command_name .. ", command delay set to " .. command_delay .. " " .. delay_units)
    if (command == "") then
	   LogWarn("PackAndQueueCommand(), command_name = " .. command_name .. ", command string is empty - exiting PackAndQueueCommand()")
	   return
    end
    
    local FinalCommandHeader = gWriteFunctionCommandHeader
    if  ((command_name == "MENU") or 
	    (command_name == "CANCEL") or
	    (command_name == "UP") or 
	    (command_name == "DOWN") or 
	    (command_name == "LEFT") or 
	    (command_name == "RIGHT") or 
	    (command_name == "ENTER"))then
	   LogTrace("Change FinalCommandHeader = " .. command_name .. "-->" .. gWriteFunctionCommandHeader_RemoteKey )
	   FinalCommandHeader = gWriteFunctionCommandHeader_RemoteKey
    end
	
    local s = FinalCommandHeader .. command
    local checksum = GetChecksum(s:sub(2))
    
    local cmd, stx, etx
    if (gControlMethod == "Network") then
		stx = FinalCommandHeader
		etx = checksum		--"\r"
		cmd = stx .. command .. etx
    elseif (gControlMethod == "Serial") then
		stx = FinalCommandHeader
		etx = checksum		--"\r"
		cmd = stx .. command .. etx
    elseif (gControlMethod == "IR") then
		cmd = command
    else
		LogWarn("PackAndQueueCommand(): gControlMethod is not valid, ".. tostring(gControlMethod))
		return
    end
    gCon:QueueCommand(cmd, command_delay, delay_units, command_name)	
	
end

function PackAndQueuePriority1Command(...)
    local command_name = select(1, ...) or ""
    local command = select(2, ...) or ""
    local command_delay = select(3, ...) or tonumber(Properties["Command Delay Milliseconds"])
    local delay_units = select(4, ...) or "MILLISECONDS"
    LogTrace("PackAndQueuePriority1Command(), command_name = " .. command_name .. ", command delay set to " .. command_delay .. " " .. delay_units)
    if (command == "") then
	   LogWarn("PackAndQueuePriority1Command(), command_name = " .. command_name .. ", command string is empty - exiting PackAndQueuePriority1Command()")
	   return
    end
	
	-- TODO: pack command with any any required starting or ending characters
	
    local FinalCommandHeader = gWriteFunctionCommandHeader
    if  ((command_name == "MENU") or 
	    (command_name == "CANCEL") or
	    (command_name == "UP") or 
	    (command_name == "DOWN") or 
	    (command_name == "LEFT") or 
	    (command_name == "RIGHT") or 
	    (command_name == "ENTER"))then
	   LogTrace("Change FinalCommandHeader = " .. command_name .. "-->" .. gWriteFunctionCommandHeader_RemoteKey )
	   FinalCommandHeader = gWriteFunctionCommandHeader_RemoteKey
    end
	
    local cmd, stx, etx
    if (gControlMethod == "Network") then
		stx = FinalCommandHeader
		etx = checksum		--"\r"
		cmd = stx .. command .. etx
    elseif (gControlMethod == "Serial") then
		stx = FinalCommandHeader
		etx = checksum		--"\r"
		cmd = stx .. command .. etx
    elseif (gControlMethod == "IR") then
		cmd = command
    else
		LogWarn("PackAndQueuePriority1Command(): gControlMethod is not valid, ".. tostring(gControlMethod))
		return
    end
    gCon:QueuePriority1Command(cmd, command_delay, delay_units, command_name)	
	
end

function PackAndQueueReadCommand(...)
    local command_name = select(1, ...) or ""
    local command = select(2, ...) or ""
    local command_delay = select(3, ...) or tonumber(Properties["Command Delay Milliseconds"])
    local delay_units = select(4, ...) or "MILLISECONDS"
    LogTrace("PackAndQueueReadCommand(), command_name = " .. command_name .. ", command delay set to " .. command_delay .. " " .. delay_units)
    if (command == "") then
	   LogWarn("PackAndQueueReadCommand(), command_name = " .. command_name .. ", command string is empty - exiting PackAndQueueReadCommand()")
	   return
    end
	
    local s = gReadFunctionCommandHeader .. command
    local checksum = GetChecksum(s:sub(2))
    
    local cmd, stx, etx
    if (gControlMethod == "Network") then
		stx = gReadFunctionCommandHeader
		etx = checksum		--"\r"
		cmd = stx .. command .. etx
    elseif (gControlMethod == "Serial") then
		stx = gReadFunctionCommandHeader
		etx = checksum		--"\r"
		cmd = stx .. command .. etx
    elseif (gControlMethod == "IR") then
		cmd = command
    else
		LogWarn("PackAndQueueReadCommand(): gControlMethod is not valid, ".. tostring(gControlMethod))
		return
    end
    gCon:QueueCommand(cmd, command_delay, delay_units, command_name)	
	
end

function PackAndQueueReadPriority1Command(...)
    local command_name = select(1, ...) or ""
    local command = select(2, ...) or ""
    local command_delay = select(3, ...) or tonumber(Properties["Command Delay Milliseconds"])
    local delay_units = select(4, ...) or "MILLISECONDS"
    LogTrace("PackAndQueueReadPriority1Command(), command_name = " .. command_name .. ", command delay set to " .. command_delay .. " " .. delay_units)
    if (command == "") then
	   LogWarn("PackAndQueueReadPriority1Command(), command_name = " .. command_name .. ", command string is empty - exiting PackAndQueueReadPriority1Command()")
	   return
    end
	
    local s = gReadFunctionCommandHeader .. command
    local checksum = GetChecksum(s:sub(2))
    
    local cmd, stx, etx
    if (gControlMethod == "Network") then
		stx = gReadFunctionCommandHeader
		etx = checksum		--"\r"
		cmd = stx .. command .. etx
    elseif (gControlMethod == "Serial") then
		stx = gReadFunctionCommandHeader
		etx = checksum		--"\r"
		cmd = stx .. command .. etx
    elseif (gControlMethod == "IR") then
		cmd = command
    else
		LogWarn("PackAndQueueReadPriority1Command(): gControlMethod is not valid, ".. tostring(gControlMethod))
		return
    end
    gCon:QueuePriority1Command(cmd, command_delay, delay_units, command_name)	
	
end

function GetChecksum(s)
  local cs = 0
  for c in s:gmatch('.') do
    cs = cs + c:byte()
  end
  return string.pack("b",cs)
end
